#!/usr/bin/env python3
#
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

if __name__ == "__main__":  # pragma: no cover
    unittest.main(module="bowler.tests", verbosity=2)
